package parser;

import java.lang.*;
import java.io.*;

public class Parse {

   lexer.Lex scanner; 
   lexer.Token current;

   public Parse (String filename) throws Exception {
       scanner = new lexer.Lex(filename);
       current = scanner.nextToken();
   }

   /********************************************************/
   /*                       Debugging                      */
   /********************************************************/
   boolean TRACE = false;
   int level = 0;

   void ENTER(String m) {
      if (TRACE) {
         for(int i=0; i<level; i++)
            System.out.print("   ");
         System.out.println("ENTER " + m + " " + current.toString());
      }
      level++;
   }

   void EXIT(String m) {
      level--;
      if (TRACE) {
         for(int i=0; i<level; i++)
            System.out.print("   ");
         System.out.println("EXIT " + m + " " + current.toString());
      }
   }

   /********************************************************/
   /*                        Matching                      */
   /********************************************************/
   boolean lookahead(int tok) {
        return current.getKind() == tok;
   }

   boolean lookahead(lexer.TokenSet toks) {
      return toks.member(current);
   }

   void match(int expected) {
      match(new lexer.TokenSet(expected));
   }

   void match(lexer.TokenSet expected) {
      if (lookahead(expected)) {
         lexer.Token next = scanner.nextToken();
         if (TRACE) {
            for(int i=0; i<level; i++)
               System.out.print("   ");
            System.out.println("MATCHED " + lexer.Token.token2string(current.getKind()) + 
                               " NEXT " + lexer.Token.token2string(next.getKind()));
         }
         current = next;
      } else 
         error(expected);
   }

   void synch(lexer.TokenSet synchSet) {
      while (true) {
         int kind = current.getKind();
         if (kind == lexer.Token.EOF) return;
         if (synchSet.member(kind)) return;
         if (TRACE)
            System.out.println("SKIPPING " + current.toString());
         current = scanner.nextToken();
      }
   }

   void error(lexer.TokenSet expected) {
      String err = "<SYNTAX_ERROR " +
                   "pos=\"" + current.getPosition() + "\" " +
                   "\n   expected=\"" + expected.toString(false) + "\" " +
                   "\n   found=\"" + lexer.Token.token2stringPrint(current.getKind()) + "\"/>";
      System.out.println(err);
      System.exit(0);
   }

   void error(lexer.TokenSet expected, lexer.TokenSet synchSet) {
      String err = "SYNTAX ERROR LINE " +
                   current.getPosition() +
                   ": EXPECTED " + 
                   expected.toString(true) + 
                   " FOUND " +
                   lexer.Token.token2stringPrint(current.getKind());
      System.err.println(err);
      //      synch(synchSet);
   }

   /********************************************************/
   /*                FIRST and FOLLOW sets                 */
   /********************************************************/
   lexer.TokenSet decl_FIRST = new lexer.TokenSet(new int[]{
                        lexer.Token.VAR,
                        lexer.Token.CONST,
                        lexer.Token.TYPE,
                        lexer.Token.PROCEDURE,
                        lexer.Token.METHOD
                       });
   lexer.TokenSet decl_SYNCH = new lexer.TokenSet(new int[]{
                        lexer.Token.SEMICOLON
                       });
   lexer.TokenSet typedecl_FIRST = new lexer.TokenSet(new int[]{
                        lexer.Token.ARRAY,
                        lexer.Token.RECORD,
			lexer.Token.REF,
			lexer.Token.ENUM,
			lexer.Token.CLASS,
                       });
   lexer.TokenSet stat_FIRST = new lexer.TokenSet(new int[]{
                        lexer.Token.IDENT,
                        lexer.Token.IF,
                        lexer.Token.WHILE,
                        lexer.Token.FOR,
                        lexer.Token.REPEAT,
                        lexer.Token.LOOP,
                        lexer.Token.EXIT,
                        lexer.Token.READ,
                        lexer.Token.SPECIAL,
                        lexer.Token.WRITE,
                        lexer.Token.WRITELN,
                        lexer.Token.LPAREN,
                       });
   lexer.TokenSet assignORcall_FIRST = new lexer.TokenSet(new int[]{
                        lexer.Token.COLONEQ,
                        lexer.Token.LPAREN
                       });
   lexer.TokenSet stat_FOLLOW = new lexer.TokenSet(
                        lexer.Token.SEMICOLON
                       );
   lexer.TokenSet expression_FIRST = new lexer.TokenSet(new int[]{
                                  lexer.Token.NOT,
                                  lexer.Token.TRUNC,
                                  lexer.Token.FLOAT,
                                  lexer.Token.MINUS,
                                  lexer.Token.IDENT,
                                  lexer.Token.INTLIT,
                                  lexer.Token.STRINGLIT,
                                  lexer.Token.CHARLIT,
                                  lexer.Token.LPAREN
                               });
   lexer.TokenSet logical_FIRST = new lexer.TokenSet(new int[]{
                                  lexer.Token.AND,
                                  lexer.Token.OR
                               });
   lexer.TokenSet relational_FIRST = new lexer.TokenSet(new int[]{
                                  lexer.Token.EQ,
                                  lexer.Token.NE,
                                  lexer.Token.GT,
                                  lexer.Token.GE,
                                  lexer.Token.LT,
                                  lexer.Token.LE,
                                  lexer.Token.ISA
                               });
   lexer.TokenSet unary_FIRST = new lexer.TokenSet(new int[]{
                                  lexer.Token.NOT,
                                  lexer.Token.TRUNC,
                                  lexer.Token.FLOAT,
                                  lexer.Token.MINUS
                               });
   lexer.TokenSet factor_FIRST = new lexer.TokenSet(new int[]{
                                  lexer.Token.IDENT,
                                  lexer.Token.INTLIT,
                                  lexer.Token.STRINGLIT,
                                  lexer.Token.REALLIT,
                                  lexer.Token.CHARLIT,
                                  lexer.Token.LPAREN,
                                  lexer.Token.NOT,
                                  lexer.Token.TRUNC,
                                  lexer.Token.FLOAT,
                                  lexer.Token.MINUS,
                                  lexer.Token.NEW,
                               });
   lexer.TokenSet expr_FIRST = new lexer.TokenSet(
                                  lexer.Token.PLUS,
                                  lexer.Token.MINUS
                               );
   lexer.TokenSet term_FIRST = new lexer.TokenSet(
                                  lexer.Token.STAR,
                                  lexer.Token.PERCENT,
                                  lexer.Token.SLASH
                               );
   lexer.TokenSet formals_FIRST = new lexer.TokenSet(
                                  lexer.Token.IDENT,
                                  lexer.Token.VAR
                               );


   /********************************************************/
   /*                      Program                         */
   /********************************************************/
   public ast.PROGRAM program () {
      ENTER("PROGRAM");
      int pos = current.getPosition();
      match(lexer.Token.PROGRAM);
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.SEMICOLON);
      ast.DECLS d = decls();
      ast.STATS s = block();
      match(lexer.Token.PERIOD);
      match(lexer.Token.EOF);
      ast.PROGRAM p = new ast.PROGRAM(name, d, s, pos);
      EXIT("PROGRAM");
      return p;
   }

   /********************************************************/
   /*                      Statements                      */
   /********************************************************/
   ast.STATS block()  {
      ENTER("block");
      match(lexer.Token.BEGIN);
      ast.STATS s = stats();
      match(lexer.Token.END);
      EXIT("block");
      return s;
   }


   ast.STATS stats()  {
      ENTER("stats");
      ast.STATS p = new ast.STATNULL(current.getPosition());
      while (lookahead(stat_FIRST)) {
         ast.STATEMENT s = stat();
         int pos = current.getPosition();
         match(lexer.Token.SEMICOLON);
         ast.STATS r = stats();
         p = new ast.STATS(s, r, pos);
      }
      EXIT("stats");
      return p;
   }

   lexer.TokenSet stat_SYNCH = stat_FOLLOW.union(decl_FIRST);
   ast.STATEMENT stat()  {
      ast.STATEMENT s = null;
      if (lookahead(lexer.Token.IDENT)) 
         s = assignORcall();
      else if (lookahead(lexer.Token.IF)) 
         s = if_();
      else if (lookahead(lexer.Token.FOR)) 
         s = for_();
      else if (lookahead(lexer.Token.WHILE)) 
         s = while_();
      else if (lookahead(lexer.Token.REPEAT)) 
         s = repeat();
      else if (lookahead(lexer.Token.LOOP)) 
         s = loop();
      else if (lookahead(lexer.Token.EXIT)) 
         s = exit_();
      else if (lookahead(lexer.Token.READ)) 
         s = read();
      else if (lookahead(lexer.Token.SPECIAL)) 
         s = special();
      else if (lookahead(lexer.Token.WRITE)) 
         s = write();
      else if (lookahead(lexer.Token.WRITELN)) 
         s = writeln();
      else
         error(stat_FIRST);
      return s;
   }

   ast.STATEMENT assignORcall() {
      ENTER("assignORcall");
      ast.STATEMENT p = null;
      ast.DESIGNATOR d = designator();
      int pos = current.getPosition();
      if (lookahead(lexer.Token.COLONEQ)) {
         match(lexer.Token.COLONEQ);
         ast.EXPRESSION e = expression();      
         p = new ast.ASSIGN(d, e, pos);
      } else if (lookahead(lexer.Token.LPAREN)) {
         match(lexer.Token.LPAREN);
         ast.ACTUAL a = actuals();
         match(lexer.Token.RPAREN);
         p = new ast.PROCCALL(d, a, pos);
      } else
         error(assignORcall_FIRST);
      EXIT("assignORcall");
      return p;
   }

   ast.ACTUAL actuals() {
      ENTER("actuals");
      ast.ACTUAL p = null;
      if (lookahead(expression_FIRST)) {
         int pos = current.getPosition();
         ast.EXPRESSION e = expression();
         p = actuals1();
         p = new ast.ACTUAL(e, p, pos);
      } else
        p = new ast.ACTUALNULL(current.getPosition());
      EXIT("actuals");
      return p;
   }

   ast.ACTUAL actuals1() {
      ENTER("actuals1");
      ast.ACTUAL p = null;
      if (lookahead(lexer.Token.COMMA)) {
         match(lexer.Token.COMMA);
         int pos = current.getPosition();
         ast.EXPRESSION e = expression();
         p = actuals1();
         p = new ast.ACTUAL(e, p, pos);
      } else
        p = new ast.ACTUALNULL(current.getPosition());
      EXIT("actuals1");
      return p;
   }

   ast.STATEMENT if_() {
      ENTER("if");
      int pos = current.getPosition();
      match(lexer.Token.IF);
      ast.EXPRESSION e = expression();      
      match(lexer.Token.THEN);
      ast.STATS s1 = stats();
      ast.STATEMENT p = null;
      if (lookahead(lexer.Token.ELSE)) {
         match(lexer.Token.ELSE);
         ast.STATS s2 = stats();
         p = new ast.IF2(e, s1, s2, pos);
      } else      
         p = new ast.IF1(e, s1, pos);
      match(lexer.Token.ENDIF);
      EXIT("if");
      return p;
   }

   ast.STATEMENT while_() {
      ENTER("while");
      int pos = current.getPosition();
      match(lexer.Token.WHILE);
      ast.EXPRESSION e = expression();      
      match(lexer.Token.DO);
      ast.STATS s = stats();
      ast.STATEMENT p = new ast.WHILE(e, s, pos);
      match(lexer.Token.ENDDO);
      EXIT("while");
      return p;
   }

   ast.STATEMENT repeat() {
      ENTER("repeat");
      int pos = current.getPosition();
      match(lexer.Token.REPEAT);
      ast.STATS s = stats();
      match(lexer.Token.UNTIL);
      ast.EXPRESSION e = expression();      
      ast.STATEMENT p = new ast.REPEAT(e, s, pos);
      EXIT("repeat");
      return p;
   }

   ast.STATEMENT for_() {
      ENTER("for");
      int pos = current.getPosition();
      match(lexer.Token.FOR);
      String id = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.COLONEQ);
      ast.EXPRESSION from = expression();      
      match(lexer.Token.TO);
      ast.EXPRESSION to = expression();      
      ast.EXPRESSION by = byexpr();      
      match(lexer.Token.DO);
      ast.STATS s = stats();
      ast.STATEMENT p = new ast.FOR(id, from, to, by, s, pos);
      match(lexer.Token.ENDFOR);
      EXIT("for");
      return p;
   }

    ast.EXPRESSION byexpr() {
      ast.EXPRESSION by = null;
      if (lookahead(lexer.Token.BY)) {
         match(lexer.Token.BY);
         by = expression();      
      }
      return by;
    }

   ast.STATEMENT loop() {
      ENTER("loop");
      int pos = current.getPosition();
      match(lexer.Token.LOOP);
      ast.STATS s = stats();
      match(lexer.Token.ENDLOOP);
      ast.STATEMENT p = new ast.LOOP(s, pos);
      EXIT("loop");
      return p;
   }

   ast.STATEMENT exit_() {
      ENTER("exit");
      int pos = current.getPosition();
      match(lexer.Token.EXIT);
      ast.STATEMENT p = new ast.EXIT( pos);
      EXIT("exit");
      return p;
   }

   ast.STATEMENT read() {
      ENTER("read");
      int pos = current.getPosition();
      match(lexer.Token.READ);
      ast.DESIGNATOR d = designator();
      ast.STATEMENT p = new ast.READ(d, pos);
      EXIT("read");
      return p;
   }

   ast.STATEMENT write() {
      ENTER("write");
      int pos = current.getPosition();
      match(lexer.Token.WRITE);
      ast.EXPRESSION e = expression();
      ast.STATEMENT p = new ast.WRITE(e, pos);
      EXIT("write");
      return p;
   }

   ast.STATEMENT writeln() {
      ENTER("writeln");
      int pos = current.getPosition();
      match(lexer.Token.WRITELN);
      ast.STATEMENT p = new ast.WRITELN(pos);
      EXIT("writeln");
      return p;
   }

   ast.STATEMENT special() {
      ENTER("special");
      int pos = current.getPosition();
      match(lexer.Token.SPECIAL);
      String value = current.getValue();
      match(lexer.Token.STRINGLIT);
      ast.STATEMENT p = new ast.SPECIAL(value, pos);
      EXIT("special");
      return p;
   }

   /********************************************************/
   /*                     Declarations                     */
   /********************************************************/
   ast.DECLS decls() {
      ENTER("decls");
      ast.DECLS p = new ast.DECLNULL(current.getPosition());
      while (lookahead(decl_FIRST)) {
         ast.DECLARATION s = decl();
         int pos = current.getPosition();
         match(lexer.Token.SEMICOLON);
         ast.DECLS r = decls();
         p = new ast.DECLS(s, r, pos);
      }
      EXIT("decls");
      return p;
   }

   ast.DECLARATION decl() {
      ast.DECLARATION s = null;
      if (lookahead(lexer.Token.VAR)) 
         s = vardecl();
      else if (lookahead(lexer.Token.CONST)) 
         s = constdecl();
      else if (lookahead(lexer.Token.TYPE)) 
         s = typedecl();
      else if (lookahead(lexer.Token.PROCEDURE)) 
         s = procdecl();
      else
         error(decl_FIRST);
      return s;
   }
   
   ast.VARDECL vardecl() {
      int pos = current.getPosition();
      match(lexer.Token.VAR);
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.COLON);
      String type = current.getValue();
      match(lexer.Token.IDENT);
      return new ast.VARDECL(name,type,pos);
   }

   ast.CONSTDECL constdecl() {
      int pos = current.getPosition();
      match(lexer.Token.CONST);
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.COLON);
      String type = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.EQ);
      ast.EXPRESSION expr = expression();
      return new ast.CONSTDECL(name,type,expr,pos);
   }

   ast.DECLARATION typedecl() {
      match(lexer.Token.TYPE); 
      int pos = current.getPosition();
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.EQ);
      ast.DECLARATION s = null;
      if (lookahead(lexer.Token.ARRAY)) 
         s = arraydecl(name,pos);
      else if (lookahead(lexer.Token.RECORD)) 
         s = recorddecl(name,pos);
      else if (lookahead(lexer.Token.REF)) 
         s = refdecl(name,pos);
      else if (lookahead(lexer.Token.ENUM)) 
         s = enumdecl(name,pos);
      else if (lookahead(lexer.Token.CLASS)) 
	  s = classdecl(name,pos);
      else
         error(typedecl_FIRST);
      return s;
   }

   ast.ARRAYDECL arraydecl(String name, int pos) {
      match(lexer.Token.ARRAY); 
      ast.EXPRESSION expr = expression();
      match(lexer.Token.OF);  
      String type = current.getValue();
      match(lexer.Token.IDENT);
      return new ast.ARRAYDECL(name, expr, type, pos);
   }

   ast.REFDECL refdecl(String name, int pos) {
      match(lexer.Token.REF); 
      String type = current.getValue();
      match(lexer.Token.IDENT);
      return new ast.REFDECL(name, type, pos);
   }

   ast.RECORDDECL recorddecl(String name, int pos) {
      match(lexer.Token.RECORD); 
      match(lexer.Token.LBRACK); 
      ast.DECLS fields = fielddecls();
      match(lexer.Token.RBRACK); 
      return new ast.RECORDDECL(name, fields, pos);
   }

   ast.DECLS fielddecls() {
      ENTER("fielddecls");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.IDENT)) {
         int pos = current.getPosition();
         ast.FIELDDECL e = fielddecl();
         p = fielddecls1();
         p = new ast.DECLS(e, p, pos);
      } else
        p = new ast.DECLNULL(current.getPosition());
      EXIT("fielddecls");
      return p;
   }

   ast.DECLS fielddecls1() {
      ENTER("fielddecls1");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.SEMICOLON)) {
         match(lexer.Token.SEMICOLON);
         int pos = current.getPosition();
         ast.FIELDDECL e = fielddecl();
         p = fielddecls1();
         p = new ast.DECLS(e, p, pos);
      } else
         p = new ast.DECLNULL(current.getPosition());
      EXIT("fielddecls1");
      return p;
   }

   ast.FIELDDECL fielddecl() {
      int pos = current.getPosition();
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.COLON);
      String type = current.getValue();
      match(lexer.Token.IDENT);
      return new ast.FIELDDECL(name,type,pos);
   }


   ast.DECLS classfielddecls() {
      ENTER("classfielddecls");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.IDENT)) {
         int pos = current.getPosition();
         ast.CLASSFIELDDECL e = classfielddecl();
         p = classfielddecls1();
         p = new ast.DECLS(e, p, pos);
      } else
        p = new ast.DECLNULL(current.getPosition());
      EXIT("classfielddecls");
      return p;
   }

   ast.DECLS classfielddecls1() {
      ENTER("classfielddecls1");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.SEMICOLON)) {
         match(lexer.Token.SEMICOLON);
         int pos = current.getPosition();
         ast.CLASSFIELDDECL e = classfielddecl();
         p = classfielddecls1();
         p = new ast.DECLS(e, p, pos);
      } else
         p = new ast.DECLNULL(current.getPosition());
      EXIT("classfielddecls1");
      return p;
   }

   ast.CLASSFIELDDECL classfielddecl() {
      int pos = current.getPosition();
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.COLON);
      String type = current.getValue();
      match(lexer.Token.IDENT);
      return new ast.CLASSFIELDDECL(name,type,pos);
   }

   ast.DECLS methoddecls() {
      ENTER("methoddecls");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.METHOD)) {
         int pos = current.getPosition();
         ast.METHODDECL e = methoddecl();
         p = methoddecls1();
         p = new ast.DECLS(e, p, pos);
      } else
        p = new ast.DECLNULL(current.getPosition());
      EXIT("methoddecls");
      return p;
   }

   ast.DECLS methoddecls1() {
      ENTER("methoddecls1");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.SEMICOLON)) {
         match(lexer.Token.SEMICOLON);
         int pos = current.getPosition();
         ast.METHODDECL e = methoddecl();
         p = methoddecls1();
         p = new ast.DECLS(e, p, pos);
      } else
         p = new ast.DECLNULL(current.getPosition());
      EXIT("methoddecls1");
      return p;
   }

   ast.METHODDECL methoddecl() {
      int pos = current.getPosition();
      match(lexer.Token.METHOD);
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.LPAREN); 
      ast.DECLS formals = formals();
      match(lexer.Token.RPAREN); 
      match(lexer.Token.SEMICOLON); 
      ast.DECLS decls = decls();
      ast.STATS stats = block();
      return new ast.METHODDECL(name,formals,decls,stats,pos);
   }

    ast.EXTENDS extends_() {
      ast.EXTENDS d = null;
      if (lookahead(lexer.Token.EXTENDS)) {
         match(lexer.Token.EXTENDS);
         int pos = current.getPosition();
         String name = current.getValue();
         match(lexer.Token.IDENT);
          d = new ast.EXTENDS(name,pos);
      } else {
          int pos = current.getPosition();
	  d = new ast.EXTENDSNULL(pos);
      }
      return d;
    }

   ast.CLASSDECL classdecl(String name, int position) {
      match(lexer.Token.CLASS); 
      ast.EXTENDS extends_ = extends_();
      match(lexer.Token.LBRACK); 
      ast.DECLS fields = classfielddecls();
      match(lexer.Token.RBRACK); 
      match(lexer.Token.LBRACK); 
      ast.DECLS methods = methoddecls();
      match(lexer.Token.RBRACK); 
      return new ast.CLASSDECL(name,extends_,fields,methods,position);
   }

   ast.ENUMIDDECL enumiddecl() {
      int pos = current.getPosition();
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.EQ);
      ast.EXPRESSION expr = expression();      
      return new ast.ENUMIDDECL(name,expr,pos);
   }

   ast.DECLS enumdecls1() {
      ENTER("enumdecls1");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.COMMA)) {
         match(lexer.Token.COMMA);
         int pos = current.getPosition();
         ast.ENUMIDDECL e = enumiddecl();
         p = enumdecls1();
         p = new ast.DECLS(e, p, pos);
      } else
         p = new ast.DECLNULL(current.getPosition());
      EXIT("enumdecls1");
      return p;
   }

   ast.DECLS enumdecls() {
      ENTER("enumdecls");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.IDENT)) {
         int pos = current.getPosition();
         ast.ENUMIDDECL e = enumiddecl();
         p = enumdecls1();
         p = new ast.DECLS(e, p, pos);
      } else
        p = new ast.DECLNULL(current.getPosition());
      EXIT("enumdecls");
      return p;
   }

   ast.ENUMDECL enumdecl(String name, int position) {
      match(lexer.Token.ENUM); 
      match(lexer.Token.LBRACK); 
      ast.DECLS enums = enumdecls();
      match(lexer.Token.RBRACK); 
      return new ast.ENUMDECL(name, enums, position);
   }

   ast.PROCDECL procdecl() {
      int pos = current.getPosition();
      match(lexer.Token.PROCEDURE);
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.LPAREN); 
      ast.DECLS formals = formals();
      String type = current.getValue();
      match(lexer.Token.RPAREN); 
      match(lexer.Token.SEMICOLON); 
      ast.DECLS decls = decls();
      ast.STATS stats = block();
      return new ast.PROCDECL(name,formals,decls,stats,pos);
   }

   ast.DECLS formals() {
      ENTER("formals");
      ast.DECLS p = null;
      if (lookahead(formals_FIRST)) {
         int pos = current.getPosition();
         ast.FORMALDECL e = formaldecl();
         p = formals1();
         p = new ast.DECLS(e, p, pos);
      } else
        p = new ast.DECLNULL(current.getPosition());
      EXIT("formals");
      return p;
   }

   ast.DECLS formals1() {
      ENTER("formals1");
      ast.DECLS p = null;
      if (lookahead(lexer.Token.SEMICOLON)) {
         match(lexer.Token.SEMICOLON);
         int pos = current.getPosition();
         ast.FORMALDECL e = formaldecl();
         p = formals1();
         p = new ast.DECLS(e, p, pos);
      } else
        p = new ast.DECLNULL(current.getPosition());
      EXIT("formals1");
      return p;
   }

   ast.FORMALDECL formaldecl() {
      int pos = current.getPosition();
      String mode = "VAL";
      if (lookahead(lexer.Token.VAR)) {
         match(lexer.Token.VAR);
         mode = "VAR";
      }
      String name = current.getValue();
      match(lexer.Token.IDENT);
      match(lexer.Token.COLON);
      String type = current.getValue();
      match(lexer.Token.IDENT);
      return new ast.FORMALDECL(name,type,mode,pos);
   }

   /********************************************************/
   /*                     Expressions                      */
   /********************************************************/
   ast.EXPRESSION expression() {
      return expr();
   }

   ast.EXPRESSION expr() {
        ENTER("expr");
        ast.EXPRESSION left = term();
        while(lookahead(expr_FIRST)) {
           int position = current.getPosition();
           int op = current.getKind();
           match(expr_FIRST);
           ast.EXPRESSION right = term();
           left = new ast.BINARY(op, left, right, position);
        }
        EXIT("expr");
        return left;
    }

   ast.EXPRESSION term() {
        ENTER("term");
        ast.EXPRESSION left = logical();
        while(lookahead(term_FIRST)) {
           int position = current.getPosition();
           int op = current.getKind();
           match(term_FIRST);
           ast.EXPRESSION right = logical();
           left = new ast.BINARY(op, left, right, position);
        }
        EXIT("term");
        return left;
    }

   ast.EXPRESSION logical() {
        ENTER("logical");
        ast.EXPRESSION left = relational();
        while(lookahead(logical_FIRST)) {
           int position = current.getPosition();
           int op = current.getKind();
           match(logical_FIRST);
           ast.EXPRESSION right = relational();
           left = new ast.BINARY(op, left, right, position);
        }
        EXIT("logical");
        return left;
    }

   ast.EXPRESSION relational() {
        ENTER("relational");
        ast.EXPRESSION left = factor();
        while(lookahead(relational_FIRST)) {
           int position = current.getPosition();
           int op = current.getKind();
           match(relational_FIRST);
           if (op == lexer.Token.ISA) {
              String ident = current.getValue();
              match(lexer.Token.IDENT);
              left = new ast.ISA(op, left, ident, position);
              return left;
           } else {
              ast.EXPRESSION right = factor();
              left = new ast.BINARY(op, left, right, position);
           }
        }
        EXIT("relational");
        return left;
    }

   ast.EXPRESSION factor() {
        ENTER("factor");
        ast.EXPRESSION e=null;
        if (lookahead(lexer.Token.LPAREN)) {
           match(lexer.Token.LPAREN);
           e = expr();
           match(lexer.Token.RPAREN);
        } else if (lookahead(lexer.Token.IDENT)) {
           e = designator();
        } else if (lookahead(lexer.Token.NEW)) {
          int position = current.getPosition();
          match(lexer.Token.NEW);
          String ident = current.getValue();
          match(lexer.Token.IDENT);
          e = new ast.NEW(ident, position);
        } else if (lookahead(unary_FIRST)) {
           int position = current.getPosition();
           int op = current.getKind();
           match(unary_FIRST);
           ast.EXPRESSION left = factor();
           e = new ast.UNARY(op, left, position);
        } else if (lookahead(lexer.Token.INTLIT)) {
          String value = current.getValue();
          int pos = current.getPosition();
          match(lexer.Token.INTLIT);
          e = new ast.INTLIT(value,pos);
        } else if (lookahead(lexer.Token.REALLIT)) {
          String value = current.getValue();
          int pos = current.getPosition();
          match(lexer.Token.REALLIT);
          e = new ast.REALLIT(value,pos);
        } else if (lookahead(lexer.Token.STRINGLIT)) {
          String value = current.getValue();
          int pos = current.getPosition();
          match(lexer.Token.STRINGLIT);
          e = new ast.STRINGLIT(value,pos);
        } else if (lookahead(lexer.Token.CHARLIT)) {
          String value = current.getValue();
          int pos = current.getPosition();
          match(lexer.Token.CHARLIT);
          e = new ast.CHARLIT(value,pos);
        } else
           error(factor_FIRST);
        EXIT("factor");
        return e;
   }


   /********************************************************/
   /*                     Designator                       */
   /********************************************************/
   ast.DESIGNATOR designator()  {
      ENTER("designator");
      int pos = current.getPosition();
      String ident = current.getValue();
      match(lexer.Token.IDENT);
      ast.DESIGNATOR d  = null;
      if (lookahead(lexer.Token.BANG)) {
         match(lexer.Token.BANG);
         String enumId = current.getValue();
         match(lexer.Token.IDENT);
         d = new ast.ENUMREF(ident, enumId, pos);
      } else {
         ast.DESIGNATOR rest = designator1();
         d = new ast.VARREF(ident, rest, pos);
      }
      EXIT("designator");
      return d;
   }

   ast.DESIGNATOR designator1()  {
      ENTER("designator1");
      int pos = current.getPosition();
      ast.DESIGNATOR p = null;
      if (lookahead(lexer.Token.LBRACK)) {
         match(lexer.Token.LBRACK);
         ast.EXPRESSION expr = expression();
         match(lexer.Token.RBRACK);
         ast.DESIGNATOR rest = designator1();
         p = new ast.INDEX(expr, rest, pos);
      } else if (lookahead(lexer.Token.CARET)) {
         match(lexer.Token.CARET);
         ast.DESIGNATOR rest = designator1();
         p = new ast.DEREF(rest, pos);
      } else if (lookahead(lexer.Token.PERIOD)) {
         match(lexer.Token.PERIOD);
         String ident = current.getValue();
         match(lexer.Token.IDENT);
         ast.DESIGNATOR rest = designator1();
         p = new ast.FIELDREF(ident, rest, pos);
      } else if (lookahead(lexer.Token.ATCHAR)) {
         match(lexer.Token.ATCHAR);
         String ident = current.getValue();
         match(lexer.Token.IDENT);
         ast.DESIGNATOR rest = designator1();
         p = new ast.CLASSREF(ident, rest, pos);
      } else if (lookahead(lexer.Token.BACKQUOTE)) {
         match(lexer.Token.BACKQUOTE);
         String ident = current.getValue();
         match(lexer.Token.IDENT);
         ast.DESIGNATOR rest = designator1();
         p = new ast.NARROW(ident, rest, pos);
      } else
	  p = new ast.DESNULL(pos);
      EXIT("designator1");
      return p;
   }



   /********************************************************/
   /*                        Main                          */
   /********************************************************/
   public static void main (String args[]) throws Exception{
      Parse parser = new Parse(args[0]);
      ast.PROGRAM e = parser.program();
      System.out.println("<block>");
      System.out.print(e.toString(0));
      System.out.println("</block>");
   }
}
